
/* PairwiseStatSig : A Program for Estimating Pairwise Statistical Significance

Developed by Ankit Agrawal and Xiaoqiu Huang

Please cite:

Ankit Agrawal, Volker Brendel and Xiaoqiu Huang, Pairwise Statistical Significance 
Versus Database Statistical Significance for Local Alignment of Protein Sequences, 
In Proceedings of International Symposium on Bioinformatics Research and Applications, 
Springer Verlag (accepted) 2008. 

Copyright (c) 2008 Ankit Agrawal, Xiaoqiu Huang
All Rights Reserved.  E-mail: {ankitag,xqhuang}@cs.iastate.edu

Permission to use, copy, modify, and distribute this software and its
documentation for educational, research and non-profit purposes, without
fee, and without a written agreement is hereby granted, provided that the
above copyright notice, this paragraph and the following three paragraphs
appear in all copies.

Permission to incorporate this software into commercial products may be
obtained from Iowa State University Research Foundation, Inc.

	      Xiaoqiu Huang
	      Department of Computer Science
	      Iowa State University
	      Ames, IA 50011
	      E-mail: xqhuang@cs.iastate.edu
	      WWW: http://www.cs.iastate.edu/~xqhuang

    Proper attribution of the author as the source of the software would
    be appreciated:

    The PairwiseStatSig program estimates the pairwise statistical significance of a pairwise local 
	alignment of two protein sequences independent of any database using censored Maximum Likelihood Fitting.

    The PairwiseStatSig program is written in C and runs under Unix and Linux systems.
    We think that the program is portable to many machines.

    The pair of sequences to be compared are stored in a single file in FASTA format.
    The first line begins with the symbol '>'
    followed by the name of the sequence. The sequence is on the remaining lines.
    
>Human-beta
VHLTPEEKSAVTALWGKVNVDEVGGEALGRLLVVYPWTQRFFESFGDLSTPDAVMGNPKV
KAHGKKVLGAFSDGLAHLDNLKGTFATLSELHCDKLHVDPENFRLLGNVLVCVLAHHFGK
EFTPPVQAAYQKVVAGVANALAHKYH


  Compiling:
	cc PairwiseStatSig.c simulate.c gap4.comp.c gap3.comp.c sim.comp.c histogram.c wisememman.c -o PairwiseStatSig -lm
  
  Running:
	 PairwiseStatSig sequencePairFilename substitutionMatrixFilename n_shuffles
*/
#include	<stdlib.h>
#include    <stdio.h>
#include	<string.h>

void StatSigForSeqPair_ML(char* seqFilename, 
							  char* ,
							  int n_shuffles, 
							  double* lambda,
							  double* K,
							  double* pwAlgnScore,
							  double* pwStatSig 
							  ); 
int main(int argc,char* argv[])
{

	char seqPairFilename[200],substitutionMatrixFilename[200];
	int n_shuffles_ML;
	if(argc<4) {
		printf("Usage: %s <sequencePairFilename> <substitutionMatrixFilename> <n_shuffles> \n",argv[0]);
		exit(0);
	}
	else {
		strcpy(seqPairFilename,argv[1]);
		strcpy(substitutionMatrixFilename,argv[2]);
		n_shuffles_ML=atoi(argv[3]);
	}
	double lambda, K, pwAlgnScore, pwStatSig;
	StatSigForSeqPair_ML(seqPairFilename, substitutionMatrixFilename, n_shuffles_ML, &lambda, &K, &pwAlgnScore, &pwStatSig);
	printf("***************************************************\n");
	printf("\tPAIRWISE STATISTICAL SIGNIFICANCE\n");
	printf("\t---------------------------------");
	printf("\nPairwise Alignment Score         : %e",pwAlgnScore);
	printf("\nPairwise Statistical Significance: %e\n",pwStatSig);
	printf("***************************************************\n");

	return 0;
}
