/* 

             by

Ankit Agrawal & Xiaoqiu Huang
Department of Computer Science
Iowa State University
Ames, IA 50011
E-mail: {ankitag,xqhuang}@cs.iastate.edu
*/

#include   <stdio.h>
#include   <time.h>
#include   <math.h>
#include	<stdlib.h>
#include "histogram.h"
#define  BINNO     100		/* no. of bins in each dimension */

int NO_CENSORING = 0;
int CENSOR_LEFT_OF_PEAK = 1;
int CENSOR_LEFT_OF_GIVEN_CUTOFF = 2;

static int  nonbase = 50;	/* integer code of boundary char; 5 < nonbase < 128 */
static long queryblocklen = 26;	/* recommended query block size in Mb */
static long databblocklen = 22;	/* recommended query block size in Mb */

int no_entry=0;/*  10000	/* no. of entries in the prob distribution */
int  high=0;/* = no_entry*0.05;	/* high frequency cutoff */
int  low=0;/* = no_entry*0.0001;	/* low frequency cutoff */

  
/*static int  lowscore = 200;	/* a low bound on the score of a selected pair */
static int  lowscore = 0;	/* a low bound on the score of a selected pair */

/*static int  scorerange = 100;	/* a score range for the low bound */
static int  scorerange = 20;	/* a score range for the low bound */
static int lumpsize = 20;		/*for alocating histogram*/
static int  highscore = 3000;   /* upbound on scores of pairs of shuffled sequences */
/*static int  highscore = 3000;   /* upbound on scores of pairs of shuffled sequences */

static int  binlen = 30;	/* bin length size */
static int  pairsizeonly = 1;	/* compute the largest no. of pairs in a bin only */ 
static int  dorandpairs = 1;	/* compute the prob dist for unrelated pairs */
static double probcutoff = 1e-9;	/* prob cutoff */
static double problow = 1e-30;	/* low prob cutoff */

static int    di[128];		/* DNA letter to integer code */
static char   btab_date[100];
static char   *program;         /* program name */

static char *name2;	/* names of sequence files */

typedef struct BLOCKRECTYPE
        {  long  seqlen;	/* current block sequence length */
           long  maxseqlen;	/* largest block sequence length */
           long  namelen;	/* current block name length */
           long  maxnamelen;	/* largest block name length */
           long  num;   	/* current block sequence number */
           long  maxnum;   	/* largest block sequence number */
	}        blocktype;	/* type for the block */

typedef struct SEQRECTYPE
        {  char   *seq;		/* base sequence */
	   long   seqlen;       /* its length */
	   long   maxseqlen;	/* max length for seq */
	   char   *name;	/* sequence name */
	   long   namelen;	/* name length */
	   long   maxnamelen;	/* max length for name */
	}         seqtype;	/* type for the current sequence */

typedef struct COMPOSITETYPE
        {  char   *combase;	/* composite sequence */
  unsigned char   *comcode;	/* its integer code */
	   long   comlen;       /* their length */
           char   *comname;	/* composite sequence names */
	   long   comnamelen;   /* its length */
	   long   maxseqlen;	/* max length for seq */
	   long   comnum;	/* number of sequences */
	}         comtype;	/* type for the composite sequence */

typedef struct LOCATIONTYPE
        {  long   start;        /* start position of the sequence in comseq */
	   long   end;		/* end position of the sequence in comseq */
	   long   length;	/* its length */
	   long   nameind;	/* start position of the name in comname */
	}         loctype;	/* type for sequence positions in comseq */

static FILE  *infop;            /* file for trace information */

#define  LineLen   900  /* max length of a line in the parameter file */
extern int  nom;	/* number of substitution tables */
typedef struct MATPEN
  { int v[128][128];	/* substitution table */
    int q, r, qr;	/* gap penalties */
    int pay, w;		/* block penalty and transition penalty */
    char let;		/* output letter for each alignment column */
    char sname[LineLen + 2];	/* name of a substitution matrix */
  }  matpentype,  *matpenptr;	/* for all sets of tables and penalties */
extern matpenptr  matp;	/* all sets of substitution tables and penalties */
static int initseed;	/* initial seed */
static char *qfname;	/* query file name */

void WriteDistrToFile(FILE *, double, double, int, int);

double* globalVar_lambda_ML_censor1;
double* globalVar_K_ML_censor1;
double* globalVar_pwAlgnScore;
double* globalVar_pwStatSig;
double globalVar_PAMdist;
double globalVar_gapOpen;
double globalVar_gapExt;
double globalVar_niter;
  int   GetBestGap();
  extern int maxhoodscore;       /* maxlikelihood score */
extern int maxhooddist;        /* maxlikelihood distance */
extern int maxhoodgapopen;     /* maxlikelihood gap open penalty */
extern int maxhoodgapext;      /* maxlikelihood gap extension penalty */
extern int maxhoodgappct;      /* maxlikelihood percent gap opening */
extern int maxhoodsubpct;      /* maxlikelihood percent identity */
extern int vtable[][128];
//added on 4 Jan08 for using simdist for entire simulation
int gapopen;
int gapext;

int globalM, globalN;


//FILE *outfp;
char ccc;
int globalVar_gap4Only=0;
char* globalVar_paramFile;

void StatSigForSeqPair_ML(char* seqFilename, 
							  char* paramFilename,
							  int n_shuffles, 
							  double* lambda_ML_censor1,
							  double* K_ML_censor1,
							  double* pwAlgnScore,
							  double* pwStatSig /*(1-EXP(-Kmn*EXP(-lambda*(x))))*/
						)  
{ FILE *Qp, *ckopen();		/* pointers to query and database files */
  void *ckalloc1();		/* space-allocating function  */
  long  i, j;			/* index variables */
  int  value;			/* temporary var */
  time_t  st_time, ed_time;	/* start and end times */
  int  noarg = 4;		/* number of required arguments */
  int  ort;			/* orientation of query */

  no_entry = n_shuffles;

  /*Copy pointers to return variables to current global veriables*/
	globalVar_lambda_ML_censor1 = lambda_ML_censor1 ;
	globalVar_K_ML_censor1 = K_ML_censor1 ;
	globalVar_gap4Only = 1;
	globalVar_pwAlgnScore = pwAlgnScore;
	globalVar_pwStatSig = pwStatSig;

	
	globalVar_paramFile=paramFilename;



seqtype   cquery;	/* current query */
blocktype querybk;	/* query block */
comtype   comseq;	/* composite sequence record */
loctype   *locrec;	/* records of sequence positions in comseq */
long      tmp;
int       lastqrsym;	/* last symbol from Qp */

char      prefix[] = "comp";       /* prefix for output file names */
char      *qname;                 /* file name */
int       ind;		          /* index */


      time(&st_time);
      strftime(btab_date, sizeof(btab_date), "%b %e %Y", localtime(&st_time));

	  pairsizeonly = 0; 

	for ( i = 0; i < 128; i++ )
	   di[i] = 5; 	/* code of 'N' */
	di['a'] = di['A'] = 0;
	di['c'] = di['C'] = 1;
	di['g'] = di['G'] = 2;
	di['t'] = di['T'] = 3;
	di['n'] = di['N'] = 4;

	i = strlen(seqFilename) + strlen(prefix) + 30;
	qname = ( char * ) ckalloc1( i * sizeof(char));
	sprintf(qname, "%s.%s.info", seqFilename, prefix);
	infop = fopen(qname, "w");

        /* collect length stats about query sequences */
        Qp = ckopen(seqFilename, "r");
        CollectLenStats(Qp, queryblocklen, &querybk, &cquery);
	qfname = seqFilename;

	srand(time(NULL));

	fprintf(infop, "Bin length:    -b %d\n", binlen);
	fprintf(infop, "Do rand pairs: -d %d\n", dorandpairs);
	fprintf(infop, "Pairsize only: -e %d\n", pairsizeonly);
	fprintf(infop, "Score range:   -r %d\n", scorerange);
	fprintf(infop, "Low score:     -s %d\n", lowscore);
	fprintf(infop, "High score:    -u %d\n", highscore);
	fprintf(infop, "No. of points: -z %d\n", no_entry / 1000);

      	tmp = querybk.maxseqlen + 4;
	comseq.combase = (char * ) ckalloc1( tmp * sizeof(char) );
	comseq.comcode = (unsigned char *) ckalloc1( tmp * sizeof(unsigned char));
	comseq.maxseqlen = tmp;

      	tmp = querybk.maxnamelen + 4;
	comseq.comname = (char * ) ckalloc1( tmp * sizeof(char) );

	cquery.seq = (char * ) ckalloc1( (cquery.maxseqlen+3) * sizeof(char) );
	cquery.name = (char * ) ckalloc1( (cquery.maxnamelen+3) * sizeof(char) );
      	tmp = querybk.maxnum + 4;
	locrec = (loctype * ) ckalloc1(tmp * sizeof(loctype) );

	InitializeQuery(&querybk, &comseq);

	for ( lastqrsym = '2'; 1; )
	 { GetNextSeq(Qp, &lastqrsym, &cquery);
	   if ( querybk.num && ( cquery.seqlen == 0 ||
	        querybk.seqlen  + cquery.seqlen  + 1 > querybk.maxseqlen  ||
	        querybk.namelen + cquery.namelen + 1 > querybk.maxnamelen ||
	                             querybk.num + 1 > querybk.maxnum ) )
            {

	      AllocateArrays(cquery.maxseqlen, cquery.maxseqlen);
	      AllocateVectors(cquery.maxseqlen, cquery.maxseqlen);
		  globalN=globalM=cquery.maxseqlen;
	      CompareQueryDatabase(&comseq, &querybk, locrec);
		  
	      comseq.comlen = querybk.seqlen;
	      comseq.comnamelen = querybk.namelen;
	      comseq.comnum = querybk.num;
	      InitializeQuery(&querybk, &comseq);
	    }              /* a query block is processed */

	   if ( cquery.seqlen == 0 ) break;
	   AddSeqToComSeq(&cquery, &querybk, &comseq, locrec);
	 }
	fclose(Qp);

      time(&ed_time);
fprintf(infop,"Total running time in %f seconds\n", difftime(ed_time, st_time) );
}

/* Compute chains of segment pairs between two blocks */
CompareQueryDatabase(comptr, bkptr, locrec)
comtype   *comptr;	/* pointer to comseq */
blocktype *bkptr;	/* pointer to a block record */
loctype   locrec[];	/* position record */
{ void  *ckalloc1();
  long  j, k;		/* index */
  long  start, end;	/* start and end positions in combase */
  long  ind;		/* sequence index */
  long  qnum;		/* numbers of query and db sequences */
  int   qind;		/* index of the selected query sequence */
  int   *gap4score;	/* GAP4 alignment scores for each seq with the query */
  int   **gap3score;	/* GAP3 alignment scores for the three parameter sets */
  int   **simscore;	/* SIM alignment scores for the three parameter sets */
  char  *qname, *dname;	/* names of the query and the current sequence */
  char  *qseq, *dseq;	/* two sequences */
  int   qlen, dlen;	/* their lengths */
  int   UseGAP4();
  int   UseGAP3();
  int   UseSIM();
  int   listlen;	/* length of score lists */
  int   *indexone;	/* index of the first sequence in the pair */
  int   *indextwo;	/* index of the second sequence in the pair */
  int   sind;		/* index for the score lists */
  int score;

	qnum = bkptr->num;
	listlen = ( qnum * (qnum - 1) ) / 2;
	j = listlen + 2;
	gap4score = (int *) ckalloc1( j * sizeof(int) );
	indexone = (int *) ckalloc1( j * sizeof(int) );
	indextwo = (int *) ckalloc1( j * sizeof(int) );
	gap3score = (int **) ckalloc1( nom * sizeof(int *) );
	simscore = (int **) ckalloc1( nom * sizeof(int *) );
	for ( k = 0; k < nom; k++ )
	 { gap3score[k] = (int *) ckalloc1( j * sizeof(int) );
	   simscore[k] = (int *) ckalloc1( j * sizeof(int) );
	 }
        sind = 0;
     for ( qind = 0; qind < qnum; qind++ )
      { qname = comptr->comname + locrec[qind].nameind + 1;
	qseq = comptr->combase + locrec[qind].start; 
	qlen = locrec[qind].end - locrec[qind].start;
  	for ( ind = qind + 1; ind < qnum; ind++ )
	 { dname = comptr->comname + locrec[ind].nameind + 1;
	   dseq = comptr->combase + locrec[ind].start; 
   	   dlen = locrec[ind].end - locrec[ind].start;
	   indexone[sind] = qind;
	   indextwo[sind] = ind;
	   sind++;
	 }   /* compare the query sequence with every other query seq */
      }
     if ( sind != listlen )  fatal("Diff size");

CompSenSp(comptr, bkptr, locrec, sind,
      indexone, indextwo, gap4score, gap3score, simscore);
	free(gap4score);
	free(indexone);
	free(indextwo);
	for ( k = 0; k < nom; k++ )
	 { free(gap3score[k]);
	   free(simscore[k]);
	 }
	free(gap3score);
	free(simscore);
}

/* Produce a random sequence of given length and composition */
ShuffleSeq(seq, len)
char  seq[];		/* sequence */
int   len;		/* its length */
{ long  j, k;		/* index */
  char  tmp;		/* tmp base */
  int   one;

     for ( one = 0; one < 5; one++ )
      { for ( j = 1; j <= len; j++ )
	 { k = (long)(rand()/(RAND_MAX+1.0)*len);
	   /*k = k % len + 1;*/
	   if ( j != k )
	    { tmp = seq[j];
	      seq[j] = seq[k];
	      seq[k] = tmp;
	    }
	 }
      }
}

static double  *alignprop;	/* alignment proportions for each para. set */

CompSenSp(comptr, bkptr, locrec, sind,
indexone, indextwo, scorelist4, scorelist3, scorelistm)
comtype   *comptr;	/* pointer to comseq */
blocktype *bkptr;	/* pointer to a block record */
loctype   locrec[];	/* position record */
int       sind;		/* length of score lists */
int       *indexone;	/* index list for first sequences */
int       *indextwo;	/* index list for second sequences */
int       *scorelist4;	/* score list */
int       **scorelist3;	/* score list */
int       **scorelistm;	/* score list */
{ FILE *ckopen();	/* Sn and Sp rate file */
  int  j, k, m;

  FILE *Qp, *Rp;		/* query and database score files */
  char qname[LineLen];		/* an output file name */
  char *seq1, *seq2;	/* two sequenes in the pair */
  int  len1, len2;	/* their lengths */
  int  id1, id2;	/* their indices */
  char *name1, *name2;  /* their names */
  char *seqs;		/* source sequence */
  void  *ckalloc1();
  int   *rangap4freq;	/* GAP4 score counts of pairs of shuffled sequences */ 
  int   **rangap3freq;	/* GAP3 score counts of pairs of shuffled sequences */ 
  int   **ransimfreq;	/* SIM score counts of pairs of shuffled sequences */ 
 int   *distrgap4freq;	/* GAP4 score counts of pairs of shuffled sequences (copy)*/ 
  int   **distrgap3freq;	/* GAP3 score counts of pairs of shuffled sequences (copy)*/ 
  int   **distrsimfreq;	/* SIM score counts of pairs of shuffled sequences (copy)*/ 
 int   *list3, *listm;	/* score lists */
  int   SelectPairIndex();
  int   *pairlink;	/* linked lists of pairs in the same pair bin */
  int   lastind;	/* index of the last pair in a selected bin */
  int   allpair;	/* total number of pair */
  int   g43_4, g43_3;	/* pairwise performance counts: gap4 and gap3 */
  int   g4m_4, g4m_m;	/* pairwise performance counts: gap4 and sim */
  int   g3m_3, g3m_m;	/* pairwise performance counts: gap3 and sim */
  int   sb0_b, sb0_0;	/* pairwise performance counts: sim best and sim one */
  double gap4p, gap3p, simp, simp0;	/* probs */
  int   score;		/* similarity score */
  double tmp;
  int   UseGAP4();
  int   UseGAP3();
  int   UseSIM();
  double *Kmn, *Lam;	/* Kmn's and Lambda's for extreme value disttritions */
  double *Ser;		/* standard errors */
  double GetProb();
  double GetNewProb();
  double fac = 0.5;
  double GetSeqProb();
  double *stderrest;	/* standard errors for each parameter set */
  int    nterm;		/* no. of pairs used */
  double gap4pfit;	/* gap4 prob based on fitting */
  FILE   *fopen();
	
	sprintf(qname, "%s.scores", qfname);
        Qp = ckopen(qname, "a");
	for ( j = 0; j < sind; j++ )
	 { fprintf(Qp, "%d  %d  %d", indexone[j], indextwo[j], scorelist4[j]);
	   for ( k = 0; k < nom; k++ ) 
             fprintf(Qp, "  %d  %d", scorelist3[k][j], scorelistm[k][j]);
           fprintf(Qp, "\n");
	 }
	fclose(Qp);

	pairlink = (int *) ckalloc1( (sind + 1) * sizeof(int) );
	k = 0;
	lastind = SelectPairIndex(locrec, scorelistm[k], indexone, indextwo, pairlink, sind);
        if ( lastind < 0 )  return;
	if ( pairsizeonly ) return;

	id1 = indexone[lastind];
	id2 = indextwo[lastind];
	name1 = comptr->comname + locrec[id1].nameind + 1;
   	len1 = locrec[id1].end - locrec[id1].start;
	name2 = comptr->comname + locrec[id2].nameind + 1;
   	len2 = locrec[id2].end - locrec[id2].start;
	seq1 = (char *) ckalloc1( (len1 + 2) * sizeof(char) );
	seq2 = (char *) ckalloc1( (len2 + 2) * sizeof(char) );
	seqs = comptr->combase + locrec[id1].start; 
	for ( j = 1,k=0; j <= len1; j++ ) {
		if(seqs[j]!='?')
			seq1[++k] = seqs[j];
	}
	len1=k;		
	seqs = comptr->combase + locrec[id2].start; 
	for ( j = 1, k = len2; j <= len2; j++, k-- ) { 
	  seq2[j] = seqs[j];
	}
	j = highscore + 2;
	rangap4freq = (int *) ckalloc1( j * sizeof(int) );
	for ( m = 0; m <= highscore; m++ )
	  rangap4freq[m] = 0;
	rangap3freq = (int **) ckalloc1( nom * sizeof(int *) );
	ransimfreq = (int **) ckalloc1( nom * sizeof(int *) );
	for ( k = 0; k < nom; k++ )
	 { list3 = rangap3freq[k] = (int *) ckalloc1( j * sizeof(int) );
	   listm = ransimfreq[k] = (int *) ckalloc1( j * sizeof(int) );
	   for ( m = 0; m <= highscore; m++ )
	      list3[m] = listm[m] = 0;
	 }

	Kmn = (double *) ckalloc1( (2 * nom + 2) * sizeof(double) );
	Lam = (double *) ckalloc1( (2 * nom + 2) * sizeof(double) );
	Ser = (double *) ckalloc1( (2 * nom + 2) * sizeof(double) );
	alignprop = (double *) ckalloc1( (nom + 1) * sizeof(double) );


float *ret_mu=0, *ret_lambda=0, ret_Kmn;
ret_mu=(float*)ckalloc1(sizeof(float));
ret_lambda=(float*)ckalloc1(sizeof(float));
int gap4maxScore=0, gap3maxScore[nom], simmaxScore[nom];
for ( k = 0; k < nom; k++ )
	gap3maxScore[k] = simmaxScore[k] = 0;


struct Histogram *gap4hist;
struct Histogram *gap3hist[nom];
struct Histogram *simhist[nom];
 
gap4hist = new_Histogram(lowscore, highscore, lumpsize);
for ( k = 0; k < nom; k++ ) {
	gap3hist[k] = new_Histogram(lowscore, highscore, lumpsize);
	simhist[k] = new_Histogram(lowscore, highscore, lumpsize);
}


double dist_simdist;
int iter_simdist;



	//const LineLen=100;	
  char  alph[100+ 2], *s;		/* alphabet */
  int   size;				/* size of alphabet */
  char  *name;				/* temporary pointer */
  char  rdline[100+ 2];		/* a line */
  char  sfname[100+ 2];		/* name of a substitution file */
  char  cline[100+ 2];		/* a line of a substitution file */
  char  oflag;		/* flag for a parameter set */

  
	
	
	int v[128][128];
	int match, mismh,ms;
int i5,k5,j5;

	  { /* read a file containing alphabet and substitution weights */
	    FILE* Sp = fopen(globalVar_paramFile, "r");
	    while ( fgets(cline, 100, Sp) != NULL )
	     { sscanf(cline, "%s", alph);
	       if ( alph[0] != '#' )  break;
	     }
	    size = strlen(alph);
	    for ( i5 = 0; i5 < 128 ; i5++ )
	      for ( j5 = 0; j5 < 128 ; j5++ )
	            v[i5][j5] = 0;
	    match = mismh = 0;
	    for ( i5 = 0; i5 < size ; i5++ )
	     for ( j5 = 0; j5 <= i5 ; j5++ )
		{ (void) fscanf(Sp, "%d", &ms);
		  v[alph[i5]][alph[j5]] = v[alph[j5]][alph[i5]] = ms;
		  if ( ms > match ) match = ms;
		  if ( ms < mismh ) mismh = ms;
		}
	    fclose(Sp);
	  }




	score = UseGAP3(seq1, seq2, len1, len2, name1, name2, 10, 2, v, 999999);
	globalVar_pwAlgnScore[0]=(double)score;

  	for ( j = 0; j < no_entry; j++ )
	 { 
	
		score = UseGAP3(seq1, seq2, len1, len2, name1, name2, 10, 2, v, 999999);

			if(globalVar_gap4Only==1) {			
				for ( k = 0; k < nom; k++ )
				{ 
					AddToHistogram(simhist[k], score);
					AddToHistogram(gap3hist[k], score);		
						if ( score > highscore )  score = highscore;
					ransimfreq[k][score]++;
					rangap3freq[k][score]++;		//just added to take care of gap3
					if (score > simmaxScore[k])
						simmaxScore[k] = gap3maxScore[k] =score;
			
				}
			}


	 AddToHistogram(gap4hist, score);
	   if ( score > highscore )  score = highscore;
	   rangap4freq[score]++;
	  
	   	if (score > gap4maxScore)
		gap4maxScore = score;

	   
           ShuffleSeq(seq2, len2);

	 }


	j = highscore + 2;  
distrgap4freq = (int *) ckalloc1( j * sizeof(int) );
	for ( m = 0; m <= highscore; m++ )
	  distrgap4freq[m] = rangap4freq[m];
	distrgap3freq = (int **) ckalloc1( nom * sizeof(int *) );
	distrsimfreq = (int **) ckalloc1( nom * sizeof(int *) );
	for ( k = 0; k < nom; k++ )
	 { distrgap3freq[k] = (int *) ckalloc1( j * sizeof(int) );
	   distrsimfreq[k] = (int *) ckalloc1( j * sizeof(int) );
	   for ( m = 0; m <= highscore; m++ ) {
		   distrgap3freq[k][m]=rangap3freq[k][m];
	   		distrsimfreq[k][m]=ransimfreq[k][m]; 
	   }
	 }


	/*distrgap4freq array is now same as rangap4freq...
	rangap4freq would be modified to get cumulative distribution*/

	PerformSufAdd(rangap4freq, highscore);
	for ( k = 0; k < nom; k++ ) 
	 { PerformSufAdd(rangap3freq[k], highscore);
	   PerformSufAdd(ransimfreq[k], highscore);
	 }

int gap4scoreCutoff_forhigh=0;		//not so high score
int gap4scoreCutoff_forlow=0;		//a high score
int gap3scoreCutoff_forhigh[nom];		//not so high score
int gap3scoreCutoff_forlow[nom];		//a high score
int simscoreCutoff_forhigh[nom];		//not so high score
int simscoreCutoff_forlow[nom];		//a high score
for ( k = 0; k < nom; k++ ) {
	gap3scoreCutoff_forhigh[k]=gap3scoreCutoff_forlow[k]=simscoreCutoff_forhigh[k]=simscoreCutoff_forlow[k]=0;
}

	sprintf(qname, "%s.probdens", qfname);
        Qp = ckopen(qname, "w");


	for ( j = 0; j <= highscore; j++ ) { 
		fprintf(Qp, "%d  %d", j, rangap4freq[j]);
		if (rangap4freq[j]<=high && !gap4scoreCutoff_forhigh)
			gap4scoreCutoff_forhigh=j;
		if (rangap4freq[j]<=low && !gap4scoreCutoff_forlow)
			gap4scoreCutoff_forlow=j;

		for ( k = 0; k < nom; k++ ) {
			fprintf(Qp, "  %d  %d", rangap3freq[k][j], ransimfreq[k][j]);

			if (rangap3freq[k][j]<=high && !gap3scoreCutoff_forhigh[k])
				gap3scoreCutoff_forhigh[k]=j;
			if (rangap3freq[k][j]<=low && !gap3scoreCutoff_forlow[k])
				gap3scoreCutoff_forlow[k]=j;
			if (ransimfreq[k][j]<=high && !simscoreCutoff_forhigh[k])
				simscoreCutoff_forhigh[k]=j;
			if (ransimfreq[k][j]<=low && !simscoreCutoff_forlow[k])
				simscoreCutoff_forlow[k]=j;
		}
		fprintf(Qp, "\n");
	 }
	fclose(Qp);






	/* This code is for ML fitting*/
	char tempStri[50];

	int tti, gap4maxi, gap4z=0,gap4z1=0,gap4z2=0, simmaxi[nom], gap3maxi[nom];
	int gap4max=distrgap4freq[0];
	int gap3max[nom], gap3z[nom], gap3z1[nom], gap3z2[nom];
	int simmax[nom], simz[nom], simz1[nom], simz2[nom];
	for ( k = 0; k < nom; k++ ) {
		gap3max[k] = distrgap3freq[k][0];
		simmax[k] = distrsimfreq[k][0];
		gap3z[k]=gap3z1[k]=gap3z2[k]=gap3maxi[k]=0;
		simz[k]=simz1[k]=simz2[k]=simmaxi[k]=0;
	}

	for (tti=1;tti<highscore+1;tti++) {
		if (distrgap4freq[tti]>=gap4max) {
			gap4maxi = tti;
			gap4max=distrgap4freq[tti];
		}
	}
	for (tti=1;tti<gap4maxScore;tti++) {
		if(tti<gap4maxi) gap4z+=distrgap4freq[tti];
		if(tti<gap4scoreCutoff_forhigh) gap4z1+=distrgap4freq[tti];
		if(tti>gap4scoreCutoff_forlow) gap4z2+=distrgap4freq[tti];
	}


	for ( k = 0; k < nom; k++ ) {
		for (tti=1;tti<highscore+1;tti++) {
			if (distrgap3freq[k][tti]>=gap3max[k]) {
				gap3maxi[k] = tti;
				gap3max[k]=distrgap3freq[k][tti];
			}
			if (distrsimfreq[k][tti]>=simmax[k]) {
				simmaxi[k] = tti;
				simmax[k]=distrsimfreq[k][tti];
			}
		}
		for (tti=1;tti<gap3maxScore[k];tti++) {
			if (tti<gap3maxi[k]) gap3z[k]+=distrgap3freq[k][tti];
			if (tti<gap3scoreCutoff_forhigh[k]) gap3z1[k]+=distrgap3freq[k][tti];
			if (tti>gap3scoreCutoff_forlow[k]) gap3z2[k]+=distrgap3freq[k][tti];
		}
		for (tti=1;tti<simmaxScore[k];tti++) {
			if (tti<simmaxi[k]) simz[k]+=distrsimfreq[k][tti];
			if (tti<simscoreCutoff_forhigh[k]) simz1[k]+=distrsimfreq[k][tti];
			if (tti>simscoreCutoff_forlow[k]) simz2[k]+=distrsimfreq[k][tti];
		}
	}


int ttt;
float xAxis[highscore+1];
for (ttt=0;ttt<highscore+1;ttt++)
	xAxis[ttt]=ttt;


float tempFloat;


/*
Censoring left of peak with automatic estimation of nBelowCutoff
*/
*ret_mu=0; 
*ret_lambda=0;
ExtremeValueFitHistogram(gap4hist, CENSOR_LEFT_OF_PEAK, 0, 0, gap4maxScore, ret_mu, ret_lambda);
ret_Kmn=exp((*ret_mu) * (*ret_lambda));
tempFloat = *ret_lambda;
globalVar_lambda_ML_censor1[0] = (double)tempFloat;
globalVar_K_ML_censor1[0] = (double)ret_Kmn/(len1*len2);
globalVar_pwStatSig[0]=1-exp(-ret_Kmn*exp(-tempFloat*(globalVar_pwAlgnScore[0])));

/*	ML fitting code ends */
free_Histogram(gap4hist);
for ( k = 0; k < nom; k++ ) {
	free_Histogram(gap3hist[k]);
	free_Histogram(simhist[k]);
}

	free(rangap4freq);
	free(distrgap4freq);
	for ( k = 0; k < nom; k++ )
	 { free(rangap3freq[k]);
	   free(ransimfreq[k]);
	   free(distrgap3freq[k]);
	   free(distrsimfreq[k]);
	 }
	free(ransimfreq);
	free(rangap3freq);
	free(distrsimfreq);
	free(distrgap3freq);
	free(pairlink);
	free(seq1);
	free(seq2);
	free(Kmn);
	free(Lam);
	free(Ser);
	free(alignprop);
return;

}


/*Write the predicted distribution in the range*/
void WriteDistrToFile(fp, Kmn, lambda, low_score, high_score)
FILE *fp;
double Kmn, lambda;
int low_score, high_score;
{
	int j,idis;
	double dis;
	for ( j = low_score; j <= high_score; j++ ) { 
		dis = (1-exp(-Kmn*exp(-lambda*(j))))*no_entry;
		idis = (dis+0.5);
		fprintf(fp,"%d\t",idis);
	}
	fprintf(fp,"\n");
}




double GetProb(score, Kmn, Lam, ind)
int   score; 	/* alignment score */
double *Kmn, *Lam;	/* Kmn's and Lambda's for fitted formulas */
int    ind;		/* index for Kmn and Lam */
{ double  prob;
	
	prob = Kmn[ind] * exp( - Lam[ind] * score );
        if ( prob > 0.01 )  prob = 1.0 - exp(- prob);
        return  prob;
}

double GetNewProb(comptr, locrec, indexone, indextwo, score, lastind, Kmn, Lam)
comtype   *comptr;	/* pointer to comseq */
loctype   locrec[];	/* position record */
int       *indexone;	/* index list for first sequences */
int       *indextwo;	/* index list for second sequences */
int   score; 	/* alignment score */
int   lastind;	/* index of a pair of sequences */   
double *Kmn, *Lam;	/* Kmn's and Lambda's for fitted formulas */
{ int   value;		/* new alignment score */
  double  prob;
  char *seq1, *seq2;	/* two sequenes in the pair */
  int  len1, len2;	/* their lengths */
  int  id1, id2;	/* their indices */
  char *name1, *name2;  /* their names */
  int  GetGAP4Prop();
  int  j;
  double  K, L;		/* K and Lambda for GAP4 alignment */

	id1 = indexone[lastind];
	id2 = indextwo[lastind];
	name1 = comptr->comname + locrec[id1].nameind + 1;
   	len1 = locrec[id1].end - locrec[id1].start;
	name2 = comptr->comname + locrec[id2].nameind + 1;
   	len2 = locrec[id2].end - locrec[id2].start;
	seq1 = comptr->combase + locrec[id1].start; 
	seq2 = comptr->combase + locrec[id2].start; 
	value = GetGAP4Prop(seq1, seq2, len1, len2, name1, name2, alignprop);
        if ( value != score )  fatal("Diff score");
        K = L = 0.0;
        for ( j = 0; j < nom; j++ )
         { K += alignprop[j] * Kmn[j]; 
           L += alignprop[j] * Lam[j]; 
         }
	prob = K * exp( - L * score );
        if ( prob > 0.01 )  prob = 1.0 - exp(- prob);
        return  prob;
}

double GetSeqProb(seq1, seq2, len1, len2, name1, name2, score, Kmn, Lam)
char *seq1, *seq2;	/* two sequenes in the pair */
int  len1, len2;	/* their lengths */
char *name1, *name2;  /* their names */
int   score; 	/* alignment score */
double *Kmn, *Lam;	/* Kmn's and Lambda's for fitted formulas */
{ int   value;		/* new alignment score */
  double  prob;
  int  GetGAP4Prop();
  int  j;
  double  K, L;		/* K and Lambda for GAP4 alignment */

	value = GetGAP4Prop(seq1, seq2, len1, len2, name1, name2, alignprop);
        if ( value != score )  fatal("Diff score");
        K = L = 0.0;
        for ( j = 0; j < nom; j++ )
         { K += alignprop[j] * Kmn[j]; 
           L += alignprop[j] * Lam[j]; 
         }
	prob = K * exp( - L * score );
        if ( prob > 0.01 )  prob = 1.0 - exp(- prob);
        return  prob;
}

FitExtremeDist(ranfreq, highscore, Kmn, Lam, Ser, ind)
int    *ranfreq;	/* prob distr for gap4 scores */
int    highscore;	/* max score */
double *Kmn, *Lam;	/* Kmn's and Lambda's for fitted formulas */
double *Ser;		/* standard errors */
int    ind;		/* index for Kmn and Lam */
{ int  j, k;
  double  xsum, ysum;	/* sums of data points */
  double  xsqsum, ysqsum;	/* sums of squares of data points */
  int  n;		/* no. of data points */
  double prob;		/* prob */
  double x, y;		/* data points */
  double xysum;		/* sum of x * y */
  double Sxx, Syy, Sxy;	/* least-squares values */
  double nn;		/* double n */
  double tmp;

	n = 0;
	xsum = ysum = xsqsum = ysqsum = xysum = 0.0;
  	for ( j = 0; j <= highscore; j++ )
	 if ( ranfreq[j] <= high && ranfreq[j] >= low )   
 /*        if ( j >= 70 && j <= 90 )*/
		 
	  { n++;
	    xsum += (x = (double) j );
	    prob = (double) ranfreq[j] / (double) no_entry;
	    ysum += (y = log( prob ) );
	    xsqsum += x * x;
	    ysqsum += y * y;
	    xysum += x * y;
	  }
	if ( n < 3 ) fatal("No data point");
	nn = (double) n;
	Sxx = xsqsum - ( xsum * xsum ) / nn; 
	Syy = ysqsum - ( ysum * ysum ) / nn; 
	Sxy = xysum - ( xsum * ysum ) / nn;
	Lam[ind] = Sxy / Sxx;
	Kmn[ind] = (ysum / nn) - (xsum / nn) * Lam[ind];
	Lam[ind] = - Lam[ind];
	Kmn[ind] = exp( Kmn[ind] );
	tmp = ( Syy - (Sxy * Sxy) / Sxx ) / (double) (n - 2);
	if ( tmp <= 0.0 )  fatal("Negative value");
	Ser[ind] = sqrt( tmp );
}

PerformSufAdd(freqlist, highscore)
int  *freqlist;		/* list of score counts */
int  highscore;		/* score bound */
{  int  j;
   int  sum;
   	
	sum = 0;
	for ( j = highscore; j >= 0; j-- )
	 { sum = (freqlist[j] += sum);
	 }
}

int SelectPairIndex(locrec, scorelist, indexone, indextwo, pairlink, lsize)
loctype   locrec[];	/* position record */
int   *scorelist;	/* list of scores */
int   *indexone;	/* list of first indices */
int   *indextwo;	/* list of second indices */
int   *pairlink;	/* linked lists of pairs in bins */
int   lsize;		/* its size */
{ int  pairbin[BINNO + 1][BINNO + 1];	/* bins of pair counts */
  int  pairind[BINNO + 1][BINNO + 1];	/* bins of pair index */
  int  binno = BINNO;			/* bin size in each dimension */
  int  ind;	/* index of a bin with the largest number of pairs */
  int  id1, id2;	/* indices of two sequences in a pair */
  int  len1, len2;	/* lengths of two sequences in a pair */
  int  j, k;
  int  maxp;		/* max no. of pairs in the bins */
  FILE *ckopen();	/* Sn and Sp rate file */
  FILE *Qp;		/* query and database score files */
  double prob;		/* statistical significance of a score */
  double EstimateProb();
	
	for ( j = 0; j <= binno; j++ )
	 for ( k = 0; k <= binno; k++ )
	  { pairbin[j][k] = 0;
	    pairind[j][k] = -1;
	  }
	for ( j = 0; j < lsize; j++ )
/*	 if ( scorelist[j] >= lowscore && scorelist[j] <= lowscore + scorerange )  */
	  { id1 = indexone[j];
	    id2 = indextwo[j];
   	    len1 = locrec[id1].end - locrec[id1].start;
   	    len2 = locrec[id2].end - locrec[id2].start;
	    prob = EstimateProb(scorelist[j], len1, len2);
/*	    if ( prob > probcutoff || prob < problow )  continue;  */
	    len1 /= binlen;
	    if ( len1 < 0 )  fatal("Neg index");
	    if ( len1 > binno ) len1 = binno;
	    len2 /= binlen;
	    if ( len2 < 0 )  fatal("Neg index");
	    if ( len2 > binno ) len2 = binno;
	    pairbin[len1][len2]++;
	    pairlink[j] = pairind[len1][len2];
	    pairind[len1][len2] = j;
	  }
	maxp = 0; ind = -1;
	for ( j = 0; j <= binno; j++ )
	 for ( k = 0; k <= binno; k++ )
	  if ( pairbin[j][k] > maxp )
	   { maxp = pairbin[j][k];
	     ind = pairind[j][k];
	     if ( ind < 0 ) fatal("Neg index");
	   }
//        Qp = ckopen("PairNum", "a");
//	fprintf(Qp, "%s  %d\n", qfname, maxp);
//	fclose(Qp);
        return ind;
}

double EstimateProb(score, len1, len2)
int  score;	/* alignemnt score */
int  len1, len2; /* seq lengths */
{ double K = 0.058732;
  double L = 0.203428;
  double prob;

  prob = K * len1 * len2 * exp( - L * score );
  if ( prob > 0.01 )  prob = 1.0 - exp(- prob);
  return  prob;
}

/* It sorts the elements between positions p and r. */
QuickSortInt(list, p, r)
int  list[];	/* an array of integers */
int  p, r;	/* start and end positions in the list */
{ int  i, j;		/* index variables */
  int  sorted;		/* flag */
  int  t;		/* temp var */
  int  pivot;	        /* pivot element */
   
     if ( p >= r ) return;
     if ( p + 20 > r )
      { for ( i = p; i < r; i++ )
	 { sorted = 1;
	   for ( j = r; j > i; j--)
	     if ( list[j-1] > list[j] )
	      { t = list[j];
		list[j] = list[j-1];
		
		list[j-1] = t;
		sorted = 0;
	      }
	   if ( sorted )
	      break;
	 }
	return;
      }

    pivot = list[(p + r) / 2];
    i = p - 1;
    j = r + 1;
    while ( 1 )
     { for ( j--; list[j] > pivot; j-- )
	  ;
       for ( i++; list[i] < pivot; i++ )
	  ;
       if ( i < j )
	{ t = list[i];
	  list[i] = list[j];
	  list[j] = t;
        }
       else
	 break;
     }
    
    QuickSortInt(list, p, j);
    QuickSortInt(list, j+1, r);
}

/* Initialize the query block and comseq */
InitializeQuery(bkptr, comptr)
blocktype *bkptr;	/* pointer to a block record */
comtype   *comptr;	/* pointer to comseq */
{
	comptr->combase[0] = '#';
	comptr->comcode[0] = nonbase;
	comptr->comlen = 0;
	comptr->comname[0] = '>';
	comptr->comnamelen = 0;
	comptr->comnum = 0;
	bkptr->seqlen = 0;
	bkptr->namelen = 0;
	bkptr->num = 0;
}

/* Add the current sequence to comseq */
AddSeqToComSeq(seqptr, bkptr, comptr, locrec)
seqtype   *seqptr;	/* pointer to the sequence record */
blocktype *bkptr;       /* pointer to a block record */
comtype   *comptr;      /* pointer to comseq record */
loctype   locrec[];	/* location array */
{         char *combase;	/* comptr->combase */
unsigned  char *comcode;	/* comptr->comcode */
          char *seq;		/* seqptr->seq */
          long j, k;		/* positions */
	  int  symbol;		/* one base */
	  long comnum;		/* comptr->comnum */
          char *name;		/* seqptr->name */
          char *comname;	/* comptr->comname */
	  long len;		/* length */

	bkptr->seqlen += seqptr->seqlen + 1;
	if ( bkptr->seqlen > bkptr->maxseqlen )
	  fatal("AddSeqToComSeq: An unexpected large seqlen block");

	bkptr->namelen += seqptr->namelen + 1;
	if ( bkptr->namelen > bkptr->maxnamelen )
	  fatal("AddSeqToComSeq: An unexpected large namelen block");

	bkptr->num += 1;
	if ( bkptr->num > bkptr->maxnum )
	  fatal("AddSeqToComSeq: An unexpected large num block");

	comnum = comptr->comnum;	/* copy sequence */
	combase = comptr->combase;
	comcode = comptr->comcode;
	seq = seqptr->seq;
	len = seqptr->seqlen;
	locrec[comnum].start = k = comptr->comlen;
	for ( j = 1; j <= len; j++ )
	 { combase[++k] = symbol = seq[j];
	   comcode[k] = di[symbol];
	 }
	locrec[comnum].length = k - locrec[comnum].start;
	locrec[comnum].end = k;
	combase[++k] = '#';
	comcode[k] = nonbase;
	comptr->comlen = k;

        comname = comptr->comname;	 /* copy name */
	name = seqptr->name;
	len = seqptr->namelen;
	locrec[comnum].nameind = k = comptr->comnamelen;
	for ( j = 1; j <= len; j++ )
	  comname[++k] = (symbol = name[j]) != '\t' ? symbol : ' ';
	comname[++k] = '\0';
	comptr->comnamelen = k;
	  
	comptr->comnum += 1;
}

/* Collect length statistics of sequences in a file */ 
CollectLenStats(Ap, msize, bkptr, seqptr)
FILE  *Ap;		/* pointer to a file */
long  msize;		/* max size in Mb of blocks */
blocktype *bkptr;	/* pointer to a block record */
seqtype   *seqptr;	/* pointer to a seq record */
{ int  symbol;		/* next character */
  int  prev;		/* previous symbol */
  short heading;        /* 1: heading; 0: sequence */
  long  number;		/* number of sequences */
  long  baselen;	/* length of the current sequence */
  long  headlen;	/* length of the current sequence name */
  long  totlen;		/* total length of sequences */
  long  tothead;	/* total length of sequence names */

  long  maxbksize;	/* block size limit in bp */
  long  bknum;		/* max number of sequences in a block */
  long  bklen;		/* total length of sequences in a block */
  long  bkname;		/* total length of names in a block */
  long  seqlen;	        /* max length of sequences in the file */
  long  namelen;	/* max length of names in the file */
  long  lasthead;	/* length of the last seq name */
  long  tmp;

        /* determine the total length of sequences */
	maxbksize = msize * 1000000;
	prev = '\n';
	number = headlen = baselen = totlen = tothead = 0;
	heading = lasthead = 0;
	bknum = bkname = bklen = seqlen = namelen = 0;
	for ( ; ( symbol = getc(Ap)) != EOF; )
	  { if ( symbol == '>' && prev == '\n' )
	     { number++;
	       heading = 1;
	       tmp = totlen;
	       totlen += baselen + 1;
	       if ( totlen >= maxbksize )
	        { if ( tmp > bklen ) bklen = tmp;
		  totlen = baselen + 1;
		  if ( number > bknum ) bknum = number;
		  number = 1;
		  if ( tothead > bkname ) bkname = tothead;
		  tothead = lasthead;
		}
	       if ( baselen > seqlen ) seqlen = baselen;
	       baselen = 0;
	     }
	    if ( symbol != '\n' )
	     { if ( heading ) headlen++;
	       else baselen++;
	     }
	    if ( heading && symbol == '\n' )
	     { heading = 0;
	       tothead += (lasthead = 1 + headlen);
	       if ( headlen > namelen ) namelen =  headlen;
	       headlen = 0;
	     }
	    prev = symbol;
	  }

	if ( baselen > seqlen ) seqlen = baselen;
	seqlen += 2;
	totlen += baselen + 2;
	namelen += 2;
	if ( totlen > bklen ) bklen = totlen;
	if ( number > bknum ) bknum = number;
	if ( tothead > bkname ) bkname = tothead;

        bkptr->maxnum = bknum;
        bkptr->maxseqlen = bklen;
        bkptr->maxnamelen = bkname;
	seqptr->maxseqlen = seqlen;
	seqptr->maxnamelen = namelen;
 	rewind(Ap);
}

/* Obtain the next sequence in given orientation */
GetNextSeq(Sp, lastsymp, seqptr)
FILE *Sp;		/* sequence file */
int  *lastsymp;		/* the last symbol read from the sequence file */
seqtype   *seqptr;	/* pointer to a seq record */
{ int   symbol;		/* the next character */
  long  cseqlen;	/* length of the current sequence */
  long  cnamelen;	/* length of the current sequence's name */
  char  *cdirseq;	/* the current sequence */
  char  *cseqname;	/* name of the current sequence */

	if ( *lastsymp == '>' ) /* read next sequence */
	  symbol = '>';
	else
	  while ( (symbol = getc(Sp) ) != EOF && symbol != '>' )
	     ;
        if ( symbol == EOF )
	 { seqptr->seqlen = 0;
	   return;
	 }

	cdirseq = seqptr->seq;
	cseqname = seqptr->name;
	cseqname[0] = '>';
	for ( cnamelen = 0; (symbol = getc(Sp) ) != EOF && symbol != '\n'; )
	   cseqname[++cnamelen] = symbol;
        if ( symbol == EOF ) fatal("GetNextSeq: Unexpected seq file end");
	if ( cnamelen > seqptr->maxnamelen )
		printf("GetNextSeq: An unexpected long name");
	    //fatal("GetNextSeq: An unexpected long name");

	for ( cseqlen = 0; (symbol = getc(Sp) ) != EOF && symbol != '>'; )
	  if ( symbol != ' ' && symbol != '\n' )
	     cdirseq[++cseqlen] = symbol;
	if ( cseqlen > seqptr->maxseqlen )
	    fatal("GetNextSeq: An unexpected long sequence");

	seqptr->seqlen = cseqlen;
	seqptr->namelen = cnamelen;
	*lastsymp = symbol;
	return;
}

/* lib.c - library of C procedures. */

/* fatal - print message and die */
fatal(msg)
char *msg;
{
	(void) fprintf(stderr, "%s\n", msg);
	exit(1);
}

/* fatalf - format message, print it, and die */
fatalf(msg, val)
char *msg, *val;
{
	(void) fprintf(stderr, msg, val);
	(void) putc('\n', stderr);
	exit(1);
}
	
/* ckopen - open file; check for success */
FILE *ckopen(name, mode)
char *name, *mode;
{
	FILE *fopen(), *fp;

	if ((fp = fopen(name, mode)) == NULL)
		if ((fp = fopen(name, mode)) == NULL)
			if ((fp = fopen(name, mode)) == NULL)
				if ((fp = fopen(name, mode)) == NULL)
					if ((fp = fopen(name, mode)) == NULL)
						fatalf("In comp.rose6.c: Cannot open %s.", name);
	return(fp);
}

/* ckalloc1 - allocate space; check for success */
void *ckalloc1(amount)
size_t  amount;
{
	void *malloc(), *p;

	if ( (p = malloc(amount) ) == NULL)
		fatal("Ran out of memory.");
	return (p);
}




